/* ======================================================================
 *
 * HH   HH         vr_debug.h
 * HH   HH         Author(s): Ed Peters, Bill Sherman
 * HHHHHHH         Created: June 4, 1998
 * HH   HH         Last Modified: February 18, 2005 (slight mods on 04/07/21)
 * HH   HH
 *
 * Header file for FreeVR debugging and error output.  It seems
 * convenient to be able to distinguish between debugging and error
 * messages, so that debugging messages could be turned off without
 * having to turn off error messages.  For the most part, they will
 * both be printed on stderr.
 *
 * Copyright 2021, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#ifndef __VRDEBUG_H__
#define __VRDEBUG_H__

#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif


/*************************************/
/*** ANSI (~VT-100) terminal codes ***/
/* See: https://invisible-island.net/xterm/ctlseqs/ctlseqs.html */
/* Also see: man console_codes */
/* TODO: [04/07/21] consider moving these to a "vr_ansi.h" header file */
#define	ANSI_CSI		"\e["			/* The opening ASCI-CSI sequence '\e' = '\033' */

#define TTY_GETSCREENSIZE	"\e[18t"		/* ANSI-CSI XTWINOPS -- dtterm/xterm Window manipulation -- 18 is request screen size in chars */
#define	TTY_GETCURSORLOC	"\e[6n"			/* ANSI-CSI DSR -- '6' is Report Cursor Pos (CPR) as 'CSI r ; c R' */

#define	ANSI_CURSORLOC_SAVE	"\e7"			/* ANSI-DEC DECSC -- see also: SCP */
#define ANSI_CURSORLOC_RETURN	"\e8"			/* ANSI-DEC DECRC -- see also: RCP */
#define TTY_CURSORLOC_NEXTLINE	"\e[B"			/* ANSI-CSI CUD -- Cursor to next line */
#define TTY_CURSORLOC_NLINE	"\e[%dE"		/* ANSI-CSI CNL -- N lines down */
#define ANSI_CLEARTOEOS		"\e[J"			/* ANSI-CSI ED (Erase in Display) */
#define TTY_CLEARSCREEN		"\e[2J"			/* ANSI-CSI ED (Erase in Display) -- 0: erases down; 1: erases up; 2 erases all */
#define TTY_CLEARTORIGHT	"\e[0K"			/* ANSI-CSI EL (Erase in Line) -- 0: to right; 1: to left; 2: all */
#define TTY_CLEARTOLEFT		"\e[1K"			/* ANSI-CSI EL (Erase in Line) -- 0: to right; 1: to left; 2: all */
#define	TTY_CLEARNCHARS		"\e[%dX"		/* ANSI-CSI ECH (Erase %d Characters) */
#define TTY_REPEATCHAR		"\e[%db"		/* ANSI-CSI REP (Repeat) -- repeats the previously rendered character %d times */
//#define ANSI_MOVETOHOME		"\e[H"			/* ANSI-CSI CUP (Cursor Position) -- Home */
//#define ANSI_MOVETOEND		"\e[F"			/* ANSI-CSI CPL (Cursor Preceding Line) -- End */
#define ANSI_MOVETOUPPERLEFT	"\e[0;0H"		/* ANSI-CSI CUP (Cursor Position) -- Home */
#define ANSI_MOVETOROWCOL	"\e[%d;%dH"		/* ANSI-CSI CUP (Cursor Position) */
#define TTY_RECTFILL		"\e[%d;10;10;40;40$x"	/* ANSI-CSI DECFRA (Fill Rectangular Area) -- VT420 command that works on XTerm */
#define TTY_RECTCLEAR		"\e[10;10;40;40$z"	/* ANSI-CSI DECSERA (Erase Rectangular Area) -- VT420 command that works on XTerm */
#define TTY_RECTCLEARSEL	"\e[10;10;40;40${"	/* ANSI-CSI DECSERA (Selective Erase Rectangular Area) -- VT420 command that works on XTerm */

#define TTY_CURSORSHOW		"\e[?25h"		/* ANSI-CSI DECSET -- DECTCEM (Show cursor) */
#define TTY_CURSORHIDE		"\e[?25l"		/* ANSI-CSI DECRST -- DECTCEM (Hide cursor) */
#define ANSI_SETMARGIN_TOPBOT	"\e[%d;%dr"		/* ANSI-CSI DECSTBM (Set Scrolling Region) -- restricts text into the given lines */
#define ANSI_SETMARGIN_TOPBOT_OFF	"\e[r"		/* ANSI-CSI DECSTBM (Set Scrolling Region) -- setting to <null> turns off margins */
#define ANSI_MARGINENABLE	"\e[?69h"		/* ANSI-CSI DECRST -- DECLRMM (Disable side margins) [VT420/XTerm only] */
#define ANSI_MARGINDISABLE	"\e[?69l"		/* ANSI-CSI DECSET -- DECLRMM (Enable side margins) [VT420/XTerm only] */
#define TTY_SETMARGIN_SIDES	"\e[%d;%ds"		/* ANSI-CSI DECSLRM (Set left and right margins) -- restricts text into the given columns */
#define TTY_FILLWITHE		"\e#8"			/* ANSI-DEC DECALN -- good for testing */
#define TTY_RESET		"\e[!p"			/* ANSI-CSI DECSTR -- Soft-reset the terminal to initial state */
#define TTY_FULLRESET		"\ec"			/* ANSI RIS (Full Reset) */
#define TTY_FONTCHARSET		"\e(%d"			/* 0 = special graphics; B = Normal ASCII */

#define ANSI_TEXTNORM		"\e[0m"			/* ANSI-CSI SGR (Character Attributes) */
#define ANSI_TEXTBOLD		"\e[1m"
#define ANSI_TEXTUNDER		"\e[4m"
#define ANSI_TEXTREV		"\e[7m"
#define ANSI_TEXTNORMAL		"\e[22m"
#define ANSI_TEXTNOUNDER	"\e[24m"

#define ANSI_TEXTBLACK		"\e[30m"
#define ANSI_TEXTRED		"\e[31m"
#define ANSI_TEXTGREEN		"\e[32m"
#define ANSI_TEXTYELLOW		"\e[33m"
#define ANSI_TEXTBLUE		"\e[34m"
#define ANSI_TEXTMAGENTA	"\e[35m"
#define ANSI_TEXTCYAN		"\e[36m"
#define ANSI_TEXTWHITE		"\e[37m"
#define ANSI_TEXTSET		"\e[%dm"

#define ANSI_SETBACKNORM	"\e[49m"
#define ANSI_SETBACKYELLOW	"\e[103m"
#define ANSI_SETBACKMAGENTA	"\e[105m"
#define ANSI_SETBACKCYAN	"\e[106m"

#if 1 /* delete me (in favor of above) */
/*** These defines are for modifying the color/mode ***/
/***   of vt100 (ie. ansi) text output.             ***/
#  define	NORM_TEXT	ANSI_TEXTNORM
#  define	BOLD_TEXT	ANSI_TEXTBOLD
#  define	REVERSE_TEXT	ANSI_TEXTREV
#  define	RED_TEXT	ANSI_TEXTRED
#  define	SET_TEXT	ANSI_TEXTSET
#  define	ANSI_MOVE	ANSI_MOVETOROWCOL
#endif


/*** These defines are for setting the debug-level of     ***/
/***   various bits of information helpful for debugging. ***/

#define	ALWAYS_DBGLVL		   1
#define	AALWAYS_DBGLVL		   2
#define	COMMON_DBGLVL		  40
#define	DEFAULT_DBGLVL		  50		/* DEFAULT! */
#define	SELDOM_DBGLVL		 100
#define	RARE_DBGLVL		 200
#define	ALMOSTNEVER_DBGLVL	1000

/* do by default */
#define	CONFIG_ERROR_DBGLVL	   3
#define ERROR_DBGLVL		   5
#define	SPAWN_DBGLVL		  20

/* less often than default */
#define	CONFIG_WARN_DBGLVL	  55		/* was 10 */
#define PERFORMER_DBGLVL	  60
#define PERFORMER_DETAIL_DBGLVL	  60 + 500
#define GLX_DBGLVL		  60
#define	VARIABLE_DBGLVL		  65
#define	PARSE_DBGLVL		  70
#define PARSE_DETAIL_DBGLVL	 190
#define SELFCTRL_DBGLVL		 105
#define	TRACE_DBGLVL		 125

/* less often than seldom */
#define	MAG_DBGLVL		 150
#define	TRACKD_DBGLVL		 151
#define	STATIC_DBGLVL		 152 + 500  /* gives minutiae that should be more rare than rare */
#define	XWIN_DBGLVL		 153
#define	PINCHG_DBGLVL		 154
#define	ASCFOB_DBGLVL		 155
#define	FASTRK_DBGLVL		 156
#define	SHMEMD_DBGLVL		 157
#define	VRUIDD_DBGLVL		 158
#define	VRPN_DBGLVL		 159
#define	INPUT_DBGLVL		 180
#define	CONFIG_DBGLVL		 185
#define	CALLBACK_DBGLVL		 195
#define	CALLBACK_DETAIL_DBGLVL	 195 + 500
#define	BARRIER_DBGLVL		 196

/* less often than rare */
#define SERIAL_DBGLVL		 250
#define	MATH_DBGLVL		 260
#define OBJSEARCH_DBGLVL	 270
#define TRACKMEM_DBGLVL		 296
#define TRACKLOCKS_DBGLVL	 297


/* This structure is really a generic struct to link strings with */
/*   integer values, used here for debug names and debug level    */
/*   values, but it could be used elsewhere too.                  */
typedef	struct {
		int	code;		/* the debug value */
		char	*string;	/* the debug value's name */
	} vrCodeString;

#if defined(TEST_APP) || defined(CAVE)
/*=============================================================*/
/* Some defines specifically for compiling the test version of */
/*   some input devices -- mostly to keep the compiler happy.  */
#  define	vrDbgDo(n)	0
#  define	vrPrintf	printf
#  define	vrDbgPrintf	printf
#  define	vrErrPrintf	printf
#  define	vrFprintf	fprintf
#  define	vrAtoI		atoi
#  ifdef __linux
#    define	vrDbgPrintfN(...)/* print nothing */;
#  else
#    define	vrDbgPrintfN	/* print nothing */;
#  endif
#endif

#if !defined(vrPrintf)		/* ie. compiling something other than FreeVR */
#  define real_printf printf
#  define printf vrPrintf
void vrPrintf(char *fmt,...);
#endif
#if !defined(vrFprintf)		/* ie. compiling something other than FreeVR */
#  define real_fprintf fprintf
#  define fprintf vrFprintf
void vrFprintf(FILE *fp, char *fmt,...);
#endif
#if !defined(vrErrPrintf)
	void	vrErrPrintf(char *fmt,...);
#endif


#if !defined(vrDbgDo)
int	vrDbgDo(int debug_level);
#endif

#define vrTrace(proc, msg) \
		vrDbgPrintfN(TRACE_DBGLVL, "(%s::%d) %s -> %s\n", __FILE__, __LINE__, proc, msg)

#define vrDbg(msg) \
		vrDbgPrintfN(DEFAULT_DBGLVL, "(%s:%d) " BOLD_TEXT "%s\n" NORM_TEXT, __FILE__, __LINE__, msg)
#define vrErr(msg) vrErrPrintf("(%s:%d) " RED_TEXT "%s\n" NORM_TEXT, __FILE__, __LINE__, msg)

#if !defined(vrDbgPrintfN)
void	vrDbgPrintfN(int debug_level, char *fmt,...);
#endif
#if !defined(vrDbgPrintf)
void	vrDbgPrintf(char *fmt,...);
#endif
void	vrMsgPrintf(char *fmt,...);
void	vrDbgFlush();
void	vrNormPrintf(char *fmt,...);

void	vrDbgInfo();
int	vrDbgStringValue(char *debug_name);
char	*vrDbgValueString(int debug_code);

void	vrFatalError(const char *msg);

char	*vrSigName(int);


#ifdef __cplusplus
}
#endif

#endif

